; See http://en.wikipedia.org/wiki/Rubik's_Cube

; A cube consists of n*n*n equal pieces arranged in a 3-dim structure. n is a measure of the size of the cube.
; A piece is a (small) cube with 6 different colored sides. The sides have the same size and the angle between
; two adjacent sides is 90 degrees.

; The classic Rubik's Cube is a 3*3*3 cube build of 27 pieces.

; The layers of a cube, which are parallel to the sides of the cube, can be rotated independently of each other
; at 90 degrees turns. The number of layers in a cube is equal to 3*n. When a layer of the cube is rotated, the
; pieces in this layer gets a new position in the structure of the cube.

; For a corner piece you can see 3 sides. For a piece common to 2 sides of the cube you can see 2 sides. For
; the rest of the visible pieces you can see 1 side. If n >= 3 there will be a number of pieces you can't see
; at all because they are completely inside the cube. This number is equal to (n-2)^3.

; Although a layer of the cube is rotated, it's always the same sides of a piece you see. For pieces with 1, 2
; or 3 visible sides it's always the same sides you see. It is only the location of the piece in the cube that
; is changed.

; Six different cubes
; Pocket Cube:      2*2*2 = 2^3 =   8 pieces
; Rubik's Cube:     3*3*3 = 3^3 =  27 pieces,   1 invisible
; Rubik's Revenge:  4*4*4 = 4^3 =  64 pieces,   8 invisible
; Professor's Cube: 5*5*5 = 5^3 = 125 pieces,  27 invisible
; V-Cube 6:         6*6*6 = 6^3 = 216 pieces,  64 invisible
; V-Cube 7:         7*7*7 = 7^3 = 343 pieces, 125 invisible

#include-once

#include <Color.au3>
#include <GDIPlus.au3>

#include "Utilities.au3"

; --- Cube definitions ---

Global $iSize = 3, $iNewSize = $iSize			; $iSize is the size of the cube
Global $aPieces[$iSize][$iSize][$iSize]		; $aPieces is a 3-dim array of pieces that creates a cube
																					; Remark that $iSize equals the side length of the cube if each piece have side length 1
; Piece and cube sides
Global Enum $RIGHT = 0, $LEFT, $TOP, $BOTTOM, $FRONT, $BACK
Global $aPieceSides[7] = [ $RIGHT, $LEFT, $TOP, $BOTTOM, $FRONT, $BACK, 1 ]
; $aPieceSides[6] = OpenGL display list number

; The sides are numbered according to the X,Y,Z-coordinate system
; The cube has "center" in (0,0,0) and l = ( side length ) / 2
; For Rubik's Cube (3*3*3) l = 1.5
;  l,  0,  0 = 0    Right      X-layer parallel to the YZ-plane
; -l,  0,  0 = 1    Left       X                       YZ
;  0,  l,  0 = 2    Top        Y                       XZ
;  0, -l,  0 = 3    Bottom     Y                       XZ
;  0,  0,  l = 4    Front      Z                       XY
;  0,  0, -l = 5    Back       Z                       XY

; The X,Y,Z-coordinate system in an OpenGL window has zero point in the middle of the window with positive X-axis to the right,
; positive Y-axis upwards, and positive Z-axis out of the window towards the viewer.

; Piece colors
Global $aColors, $fNewColors = False
Global $hLogsColorsWin = 0, $aBrushes[6]
Global $aRIGHTcolor[3]  = [ 255, 000, 000 ] ; Red
Global $aLEFTcolor[3]   = [ 255, 128, 000 ] ; Orange
Global $aTOPcolor[3]    = [ 255, 255, 255 ] ; White
Global $aBOTTOMcolor[3] = [ 255, 255, 000 ] ; Yellow
Global $aFRONTcolor[3]  = [ 000, 000, 255 ] ; Blue
Global $aBACKcolor[3]   = [ 000, 255, 000 ] ; Green
Global $aSideColors[6]  = [ $aRIGHTcolor, $aLEFTcolor, $aTOPcolor, $aBOTTOMcolor, $aFRONTcolor, $aBACKcolor ]
; Remark that the index of the side colors matches the cube side enumeration above

; Piece corner color, 8 corners, 12 corner sides
Global $aCornerColor[3] = [ 000, 000, 000 ] ; Black

; --- OpenGL definitions ---

; OpenGL window background color
Global $aBackGround[3] = [ 211, 211, 211 ] ; Lightgrey

; The side length of a piece is set to 1
; A sphere with a diameter of 3.5 can contain a 2*2*2 cube
; A sphere with a diameter of 4 * 3.5 can contain a 7*7*7 cube
Global Const $dSphereDiameter = 4 * 3.5

Global Const $zNear = 1.0
Global Const $zFar  = $zNear + $dSphereDiameter

Global $dScale = 1.0, $dNewScale, $fZoom = 0

Global $iAngleX = 30, $iNewAngleX
Global $iAngleY = 45, $iNewAngleY
Global $fSetHomePos = False

; OpenGL display lists
Global Const $iFirstList = 1
Global $iStartList
Global $iNextList

; Matrices
Global $mModel = $IDENTITY_MATRIX
Global $mRotation = $IDENTITY_MATRIX

; --- Functions ---

Func NewColors()
	$aRIGHTcolor[0]  = _ColorGetRed  ( $aColors[0] )
	$aRIGHTcolor[1]  = _ColorGetGreen( $aColors[0] )
	$aRIGHTcolor[2]  = _ColorGetBlue ( $aColors[0] )
	$aLEFTcolor[0]   = _ColorGetRed  ( $aColors[1] )
	$aLEFTcolor[1]   = _ColorGetGreen( $aColors[1] )
	$aLEFTcolor[2]   = _ColorGetBlue ( $aColors[1] )
	$aTOPcolor[0]    = _ColorGetRed  ( $aColors[2] )
	$aTOPcolor[1]    = _ColorGetGreen( $aColors[2] )
	$aTOPcolor[2]    = _ColorGetBlue ( $aColors[2] )
	$aBOTTOMcolor[0] = _ColorGetRed  ( $aColors[3] )
	$aBOTTOMcolor[1] = _ColorGetGreen( $aColors[3] )
	$aBOTTOMcolor[2] = _ColorGetBlue ( $aColors[3] )
	$aFRONTcolor[0]  = _ColorGetRed  ( $aColors[4] )
	$aFRONTcolor[1]  = _ColorGetGreen( $aColors[4] )
	$aFRONTcolor[2]  = _ColorGetBlue ( $aColors[4] )
	$aBACKcolor[0]   = _ColorGetRed  ( $aColors[5] )
	$aBACKcolor[1]   = _ColorGetGreen( $aColors[5] )
	$aBACKcolor[2]   = _ColorGetBlue ( $aColors[5] )

	$aCornerColor[0] = _ColorGetRed  ( $aColors[6] )
	$aCornerColor[1] = _ColorGetGreen( $aColors[6] )
	$aCornerColor[2] = _ColorGetBlue ( $aColors[6] )
	$aBackGround[0]  = _ColorGetRed  ( $aColors[7] )
	$aBackGround[1]  = _ColorGetGreen( $aColors[7] )
	$aBackGround[2]  = _ColorGetBlue ( $aColors[7] )

	$aSideColors[0]  = $aRIGHTcolor
	$aSideColors[1]  = $aLEFTcolor
	$aSideColors[2]  = $aTOPcolor
	$aSideColors[3]  = $aBOTTOMcolor
	$aSideColors[4]  = $aFRONTcolor
	$aSideColors[5]  = $aBACKcolor

	If $hOpenGL Then _
		glClearColor( $aBackGround[0]/255, $aBackGround[1]/255, $aBackGround[2]/255, 1.0 )

	If $hLogsColorsWin Then
		For $i = 0 To 5
			_GDIPlus_BrushDispose( $aBrushes[$i] )
			$aBrushes[$i] = _GDIPlus_BrushCreateSolid( 0xFF000000 + $aColors[$i] )
		Next
	EndIf

	$fNewColors = False
EndFunc

Func SideColor( $side )
	Local $aColor = $aSideColors[$side]
	glColor3ub( $aColor[0], $aColor[1], $aColor[2] )
EndFunc

; CreatePiece() creates a piece of a cube with side lenght 1, with sides parallel to the X-, Y- and Z-axis, and with "center" in (0,0,0).
; This function is only called in two different situations: When a new cube is created it's called exactly once. See CreateCube() below.
; When a layer is rotated in a 90 degrees turn with the left mouse button (or right if you have changed the buttons in the options) it's
; called once for each piece in the layer. See TurnLayer() in TurnLayer.au3.
Func CreatePiece( ByRef $aPiece, $fNewCube = False )
	If $fNewCube Then
		glNewList( $iFirstList, $GL_COMPILE )				; Use a precompiled display list to secure good drawing performance
	Else
		If $aPiece[6] = $iFirstList Then						; Rotation of a layer in a 90 degrees turn is simulated by changing the colors of the pieces
			$aPiece[6] = $iStartList + $iNextList			; in the layer. When the colors are changed the display lists must be updated. The first time
			glNewList( $aPiece[6], $GL_COMPILE )			; the colors are changed the piece gets a new display list number.
			$iNextList += 1
		Else
			glNewList( $aPiece[6], $GL_COMPILE )			; The next time the colors are changed the display list number is reused
		EndIf
	EndIf

	; 6 sides
  glBegin( $GL_QUADS )

  ; FRONT
	SideColor( $aPiece[$FRONT] )
  glVertex3f(  0.45,  0.45, 0.5 )
  glVertex3f( -0.45,  0.45, 0.5 )
  glVertex3f( -0.45, -0.45, 0.5 )
  glVertex3f(  0.45, -0.45, 0.5 )

  ; BACK
	SideColor( $aPiece[$BACK] )
  glVertex3f(  0.45, -0.45, -0.5 )
  glVertex3f( -0.45, -0.45, -0.5 )
  glVertex3f( -0.45,  0.45, -0.5 )
  glVertex3f(  0.45,  0.45, -0.5 )

  ; RIGHT
	SideColor( $aPiece[$RIGHT] )
  glVertex3f( 0.5,  0.45, -0.45 )
  glVertex3f( 0.5,  0.45,  0.45 )
  glVertex3f( 0.5, -0.45,  0.45 )
  glVertex3f( 0.5, -0.45, -0.45 )

  ; LEFT
	SideColor( $aPiece[$LEFT] )
  glVertex3f( -0.5,  0.45,  0.45 )
  glVertex3f( -0.5,  0.45, -0.45 )
  glVertex3f( -0.5, -0.45, -0.45 )
  glVertex3f( -0.5, -0.45,  0.45 )

  ; TOP
	SideColor( $aPiece[$TOP] )
  glVertex3f(  0.45, 0.5, -0.45 )
  glVertex3f( -0.45, 0.5, -0.45 )
  glVertex3f( -0.45, 0.5,  0.45 )
  glVertex3f(  0.45, 0.5,  0.45 )

  ; BOTTOM
	SideColor( $aPiece[$BOTTOM] )
  glVertex3f(  0.45, -0.5,  0.45 )
  glVertex3f( -0.45, -0.5,  0.45 )
  glVertex3f( -0.45, -0.5, -0.45 )
  glVertex3f(  0.45, -0.5, -0.45 )

  glEnd() ; Sides

	; Corner color
	glColor3ub( $aCornerColor[0], $aCornerColor[1], $aCornerColor[2] )

  ; 8 corners
  glBegin( $GL_TRIANGLES )

  ; FRONT, RIGHT, BOTTOM
  glVertex3f( 0.45, -0.45, -0.5  )
  glVertex3f( 0.5,  -0.45, -0.45 )
  glVertex3f( 0.45, -0.5,  -0.45 )

  ; FRONT, RIGHT, TOP
  glVertex3f( 0.45, 0.45, -0.5  )
  glVertex3f( 0.5,  0.45, -0.45 )
  glVertex3f( 0.45, 0.5,  -0.45 )

  ; FRONT, LEFT, TOP
  glVertex3f( -0.45, 0.45, -0.5  )
  glVertex3f( -0.5,  0.45, -0.45 )
  glVertex3f( -0.45, 0.5,  -0.45 )

  ; FRONT, LEFT, BOTTOM
  glVertex3f( -0.45, -0.45, -0.5  )
  glVertex3f( -0.5,  -0.45, -0.45 )
  glVertex3f( -0.45, -0.5,  -0.45 )

  ; BACK, RIGHT, BOTTOM
  glVertex3f( 0.45, -0.45, 0.5  )
  glVertex3f( 0.5,  -0.45, 0.45 )
  glVertex3f( 0.45, -0.5,  0.45 )

  ; BACK, RIGHT, TOP
  glVertex3f( 0.45, 0.45, 0.5  )
  glVertex3f( 0.5,  0.45, 0.45 )
  glVertex3f( 0.45, 0.5,  0.45 )

  ; BACK, LEFT, TOP
  glVertex3f( -0.45, 0.45, 0.5  )
  glVertex3f( -0.5,  0.45, 0.45 )
  glVertex3f( -0.45, 0.5,  0.45 )

  ; BACK, LEFT, BOTTOM
  glVertex3f( -0.45, -0.45, 0.5  )
  glVertex3f( -0.5,  -0.45, 0.45 )
  glVertex3f( -0.45, -0.5,  0.45 )

  glEnd() ; Corners

  ; 12 corner sides
  glBegin( $GL_QUADS )

  ; FRONT, RIGHT
  glVertex3f( 0.45, -0.45, -0.5  )
  glVertex3f( 0.5,  -0.45, -0.45 )
  glVertex3f( 0.5,   0.45, -0.45 )
  glVertex3f( 0.45,  0.45, -0.5  )

  ; FRONT, TOP
  glVertex3f(  0.45, 0.45, -0.5  )
  glVertex3f(  0.45, 0.5,  -0.45 )
  glVertex3f( -0.45, 0.5,  -0.45 )
  glVertex3f( -0.45, 0.45, -0.5  )

  ; FRONT, LEFT
  glVertex3f( -0.45, -0.45, -0.5  )
  glVertex3f( -0.5,  -0.45, -0.45 )
  glVertex3f( -0.5,   0.45, -0.45 )
  glVertex3f( -0.45,  0.45, -0.5  )

  ; FRONT, BOTTOM
  glVertex3f(  0.45, -0.45, -0.5  )
  glVertex3f(  0.45, -0.5,  -0.45 )
  glVertex3f( -0.45, -0.5,  -0.45 )
  glVertex3f( -0.45, -0.45, -0.5  )

  ; BACK, RIGHT
  glVertex3f( 0.45, -0.45, 0.5  )
  glVertex3f( 0.5,  -0.45, 0.45 )
  glVertex3f( 0.5,   0.45, 0.45 )
  glVertex3f( 0.45,  0.45, 0.5  )

  ; BACK, TOP
  glVertex3f(  0.45, 0.45, 0.5  )
  glVertex3f(  0.45, 0.5,  0.45 )
  glVertex3f( -0.45, 0.5,  0.45 )
  glVertex3f( -0.45, 0.45, 0.5  )

  ; BACK, LEFT
  glVertex3f( -0.45, -0.45, 0.5  )
  glVertex3f( -0.5,  -0.45, 0.45 )
  glVertex3f( -0.5,   0.45, 0.45 )
  glVertex3f( -0.45,  0.45, 0.5  )

  ; BACK, BOTTOM
  glVertex3f(  0.45, -0.45, 0.5  )
  glVertex3f(  0.45, -0.5,  0.45 )
  glVertex3f( -0.45, -0.5,  0.45 )
  glVertex3f( -0.45, -0.45, 0.5  )

  ; RIGHT, BOTTOM
  glVertex3f( 0.5,  -0.45, -0.45 )
  glVertex3f( 0.5,  -0.45,  0.45 )
  glVertex3f( 0.45, -0.5,   0.45 )
  glVertex3f( 0.45, -0.5,  -0.45 )

  ; RIGHT, TOP
  glVertex3f( 0.5,  0.45, -0.45 )
  glVertex3f( 0.5,  0.45,  0.45 )
  glVertex3f( 0.45, 0.5,   0.45 )
  glVertex3f( 0.45, 0.5,  -0.45 )

  ; LEFT, BOTTOM
  glVertex3f( -0.5,  -0.45, -0.45 )
  glVertex3f( -0.5,  -0.45,  0.45 )
  glVertex3f( -0.45, -0.5,   0.45 )
  glVertex3f( -0.45, -0.5,  -0.45 )

  ; LEFT, TOP
  glVertex3f( -0.5,  0.45, -0.45 )
  glVertex3f( -0.5,  0.45,  0.45 )
  glVertex3f( -0.45, 0.5,   0.45 )
  glVertex3f( -0.45, 0.5,  -0.45 )

  glEnd() ; Corner sides

	glEndList()
EndFunc

Func NewCube( $iNewSize )
	$iAngleX = 30										; Default rotation angles
	$iAngleY = 45
	$mModel = $IDENTITY_MATRIX			; $mModel is used to save the model view matrix
	$mRotation = $IDENTITY_MATRIX		; Initially no rotations in the rotation matrix
	$fSetHomePos = False
	DeleteCube( $iSize )	; Because DeleteCube() deletes all OpenGL display lists, and
	$iSize = $iNewSize		; CreateCube() resizes the $aPieces array, NewCube() should 
	CreateCube( $iSize )	; not be executed in the middle of an OpenGL drawing cycle.
EndFunc

Func DeleteCube( $iSize )
	glDeleteLists( $iStartList, $iSize*$iSize*$iSize )
EndFunc

Func CreateCube( $iSize )
	ReDim $aPieces[$iSize][$iSize][$iSize]
	Local $aPieceSides2 = $aPieceSides
	$aPieceSides2[6] = 0
	For $x = 0 To $iSize - 1
		For $y = 0 To $iSize - 1
			For $z = 0 To $iSize - 1
				If $x > 0 And $x < $iSize - 1 And _	; A piece of the cube is completely inside the cube. Set display list
					 $y > 0 And $y < $iSize - 1 And _	; number ($aPieceSides[6]) to 0 to indicate that it should not be drawn.
					 $z > 0 And $z < $iSize - 1 Then	; For a 7*7*7 cube this is 125 pieces out of a total number of 343 pieces.
					$aPieces[$x][$y][$z] = $aPieceSides2
				Else
					$aPieces[$x][$y][$z] = $aPieceSides
				EndIf
			Next
		Next
	Next
	CreatePiece( $aPieces[0][0][0], True )		; Initially the colors on a side are NOT mixed up. One single piece is enough to
	$iStartList = glGenLists( $iSize^3 )			; create the entire cube.
	$iNextList = 0
EndFunc

; Draws the cube by drawing one piece at a time
; Remark that the drawing is done by executing a precompiled display list
Func DrawCube( $iSize )
	Local $aPieceSides
	Local $d = PieceDisplacement( $iSize )
	For $x = 0 To $iSize - 1
		For $y = 0 To $iSize - 1
			For $z = 0 To $iSize - 1
				$aPieceSides = $aPieces[$x][$y][$z]
				If $aPieceSides[6] Then											; A visible piece
					glPushMatrix()														; Save the model view matrix
						glTranslatef( $d+$x, $d+$y, $d+$z )			; Place the piece at the proper position in the cube structure
						glCallList( $aPieceSides[6] )						; To secure good performance a piece is drawn by executing a precompiled display list
					glPopMatrix()															; Restore the model view matrix
				EndIf
			Next
		Next
	Next
EndFunc

; The DisplayCube() function handles an entire OpenGL drawing cycle:
; Clears the window, saves the model view matrix, performs scaling and rotation, draws the cube, restores the matrix and swaps buffers.
; During this cycle there should not be any changes to the cube or the OpenGL matrices.
Func DisplayCube( $iSize, $iAngX, $iAngY )
	glClear( BitOR( $GL_COLOR_BUFFER_BIT, $GL_DEPTH_BUFFER_BIT ) )
	glPushMatrix()
		gluLookAt( 0,0,1+$dSphereDiameter/2, 0,0,0, 0,1,0 )
		glScalef( $dScale, $dScale, $dScale )
		glMultMatrixf( $mRotation )
		glRotatef( $iAngX, 1, 0, 0 )
		glRotatef( $iAngY, 0, 1, 0 )
		DrawCube( $iSize )
		glGetFloatv( $GL_MODELVIEW_MATRIX, $mModel )	; Save the model view matrix
	glPopMatrix()																		; Used to project the corners of the cube into screen coordinates in TurnLayer.au3
	SwapBuffers( $hDC )
EndFunc

; When the program starts up a Rubik's Cube is created and displayed in the OpenGL window. Without any rotations you would only see
; the front side of the cube. None of the other five sides would be visible. In the DisplayCube() function the cube is initially
; rotated 30 degrees ($iAngX = 30) around the X-axis and 45 degrees ($iAngY = 45) around the Y-axis:
;     glRotatef( 30, 1, 0, 0 )  -->  rotates the front side downwards and makes the top side visible
;     glRotatef( 45, 0, 1, 0 )  -->  rotates the front side to the right and makes the left side visible
; After these rotations the top, left and front sides are visible.
